#include "problem.h"
#include "node.h"

#include <iostream>
#include <fstream>
#include <math.h>

#define EPSILON 0.000000001

int getRand(int n) {
  double r = ((double)rand())/((double)(RAND_MAX+1.0));
  return (int)( ((double)n) * r  );
}

void Problem::readProbabilities(char* str) {
  vector<double> background;
  background.push_back( 0.1649 );
  background.push_back( 0.1000 );
  background.push_back( 0.0945 );
  background.push_back( 0.1076 );
  background.push_back( 0.1159 );
  background.push_back( 0.1019 );
  background.push_back( 0.0846 );
  background.push_back( 0.0692 );
  background.push_back( 0.0545 );
  background.push_back( 0.0407 );
  background.push_back( 0.0304 );
  background.push_back( 0.0198 );
  background.push_back( 0.0102 );
  background.push_back( 0.0058 );

  FILE* fp = fopen(str, "r");
  if (!fp) {
    cout << "Could not open file " << str << endl;
    exit(0);
  }

  int n(1);
  char com[200];
  com[0]='#';
  // Skip comments and header
  while (com[0]=='#') {
    fgets(com, 1000, fp);
  }
  
  // Now skip one header line (one is already skipped above)
  fgets(com, 1000, fp);
      
  int maxRes(-1);
  int res;
  float c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, c12, c13, c14;
  char name[50];
  n=1;
  while (n>0) {
    n = fscanf(fp, "%i %s %f %f %f %f %f %f %f %f %f %f %f %f %f %f", &res, name, &c1, &c2, &c3, &c4, &c5, &c6, &c7, &c8, &c9, &c10, &c11, &c12, &c13, &c14 );

    // We skip glycines with no CB
    if (n>0) {
      Node newnode;
      newnode.residue = res;
      newnode.name=new char[100];
      strcpy(newnode.name, name);
      maxRes = max(maxRes, res);
      newnode.probabilities.push_back(c1/background[0]);
      newnode.probabilities.push_back(c2/background[1]);
      newnode.probabilities.push_back(c3/background[2]);
      newnode.probabilities.push_back(c4/background[3]);
      newnode.probabilities.push_back(c5/background[4]);
      newnode.probabilities.push_back(c6/background[5]);
      newnode.probabilities.push_back(c7/background[6]);
      newnode.probabilities.push_back(c8/background[7]);
      newnode.probabilities.push_back(c9/background[8]);
      newnode.probabilities.push_back(c10/background[9]);
      newnode.probabilities.push_back(c11/background[10]);
      newnode.probabilities.push_back(c12/background[11]);
      newnode.probabilities.push_back(c13/background[12]);
      newnode.probabilities.push_back(c14/background[13]);

      double sum(0);
      for (int i=0; i < newnode.probabilities.size(); i++) {
	sum += newnode.probabilities[i];
      }

      for (int i=0; i < newnode.probabilities.size(); i++) {
	newnode.probabilities[i]/=sum;
      }

      for (int i=0; i<newnode.probabilities.size(); i++) {
	double p = newnode.probabilities[i];
	newnode.probabilities[i] = log(p); 
      }
      nodes.push_back(newnode);
    }
  }
  fclose(fp);

  residue2nodeindex.resize(maxRes+1, -1);
  for (int i=0; i<nodes.size(); i++) {
    assert(nodes[i].residue<residue2nodeindex.size());
    residue2nodeindex[nodes[i].residue] = i; 
  }
}

void Problem::readConstraints(char* str) {
  FILE* fp = fopen(str, "r");
  if (!fp) {
    cout << "Could not open file " << str << endl;
    exit(0);
  }
  
  int resA;
  int resB;
  float min, desired, maximum, weight;
  double maxweight(0);
  int n(1);
  char resNameA, resNameB;
  while (n>0) {
    n = fscanf(fp, "Constraint %c%i.CB %c%i.CB %f %f %f %f\n", &resNameA, &resA, &resNameB, &resB, &min, &desired, &maximum, &weight );
    if (n>0) {
      Constraint c;
      assert(resA < residue2nodeindex.size());
      c.resA = resA;
      c.resB = resB;
      c.resNameA = resNameA;
      c.resNameB = resNameB;
      c.nodeA = residue2nodeindex[resA];
      c.nodeB = residue2nodeindex[resB];
      assert(c.nodeB > -1);

      c.min = min;
      c.max = maximum;
      c.desired = desired;
      c.originalWeight = weight;
      maxweight = max(maxweight, c.originalWeight);
      c.inserted = false;
      c.random = getRand(1000000);
      c.sep=abs(resA-resB);

      if (c.originalWeight>EPSILON)
	constraints.push_back(c);
      else
      	noncontacts.push_back(c);
    }
  }

  fclose(fp);
  cout << "maxweight: " << maxweight << endl;

  // Normalize
  for (int i=0; i<constraints.size(); i++) {
    Constraint & cst = constraints[i];
    cst.weight = cst.originalWeight / maxweight;
  }
}

bool betterScore(double scoreAfter, double scoreBefore, int NAfter, int NBef) {
  if (scoreAfter > scoreBefore + EPSILON)
    return true;

  if (fabs(scoreAfter-scoreBefore) < EPSILON) {
    if (NAfter > NBef)
      return true;
  }
  return false;
}

bool Problem::Opt2() {
  bool improved(false);
  for (int i=0; i<constraints.size(); i++) {
    Constraint & cst_i = constraints[i];
    if (cst_i.weight < EPSILON)
      continue;
    
    for (int j=i+1; j<constraints.size(); j++) {
      Constraint & cst_j = constraints[j];
      
      if (cst_j.weight < EPSILON)
	continue;
      
      double scoreBefore = score;
      double avgSumBefore = avgSum;
      cst_i.inserted = !cst_i.inserted;
      int NBef=NConstraints;
      updateScore(cst_i, cst_i.inserted);
      cst_j.inserted = !cst_j.inserted;
      updateScore(cst_j, cst_j.inserted);
      double scoreAfter = score;
      double avgSumAfter = avgSum;
      
      if (avgSumAfter < avgSumBefore || !betterScore(scoreAfter, scoreBefore, NConstraints, NBef)) {
	cst_i.inserted = !cst_i.inserted;
	updateScore(cst_i, cst_i.inserted);
	
	cst_j.inserted = !cst_j.inserted;
	updateScore(cst_j, cst_j.inserted);
      }
      else {
	improved=true;
      }
    }
  }
  return improved;
}

bool Problem::Opt() {
  // Inserting
  for (int i=0; i<constraints.size(); i++) {
    Constraint & cst = constraints[i];
    if (cst.weight<EPSILON)
      continue;
    
    if (!cst.inserted) {
      double scoreBefore = score;
      cst.inserted = true;
      int NBef=NConstraints;
      updateScore(cst, true);
      double scoreAfter = score;
      if (!betterScore(scoreAfter, scoreBefore, NConstraints, NBef)) {
	cst.inserted = false;
	updateScore(cst, false);
      }
      else {
	return true;
      }
    }
  }
  
  // Removing
  for (int i=constraints.size()-1; i>=0; i--) {
    Constraint & cst = constraints[i];
    
    if (cst.weight<EPSILON)
      continue;
    
    if (cst.inserted) {
      double scoreBefore = score;
      cst.inserted = false;
      int NBef=NConstraints;
      updateScore(cst, false);
      double scoreAfter = score;
      
      if (!betterScore(scoreAfter, scoreBefore, NConstraints, NBef)) {
	cst.inserted = true;
	updateScore(cst, true);
      }
      else {
	return true;
      }
    }
  }
  return false;
}

void Problem::buildInitialSolution() {
  for (int i=0; i<nodes.size(); i++) {
    Node & n = nodes[i];
    double wm = n.getWeightedMean();
    double needed = wm-(double)n.numNeighbours;
    needed/=Node::contactScale;
    for (int j=0; j<constraints.size() && needed>0; j++) {
      Constraint & cst = constraints[j];
      if (cst.nodeA==i || cst.nodeB==i) {
	if (!cst.inserted) {
	  cst.inserted = true;
	  updateScore(cst, true);
	  needed--;
	}
      }
    }
  }
}

void Problem::Optimize() {
  Constraint::sortType=true;
  sort(constraints.begin(), constraints.end());

  for (int i=0; i<constraints.size(); i++) {
    constraints[i].inserted=false;
  }

  scoreFromScratch();
  buildInitialSolution();

  scoreFromScratch();

  bool improved(true);
  while (improved) {
    improved = Opt();
  }

  improved = true;
  while (improved) {
    improved = Opt2();
    if (Opt())
      improved=true;
  }

  score = scoreFromScratch(true);
  writeOutput("optimized");

  for (int i=0; i<constraints.size(); i++) {
    constraints[i].inserted=false;
  }
  
  Constraint::sortType=true;
  sort(constraints.begin(), constraints.end());
  
  for (int i=0; i<NConstraints; i++) {
    constraints[i].inserted=true;
  }
  
  writeOutput("topweight");

  for (int i=0; i<constraints.size(); i++) {
    constraints[i].inserted=false;
  }
    
  Constraint::sortType=false;
  sort(constraints.begin(), constraints.end());
  
  int ins(0);
  int i(0);
  while (ins<NConstraints) {
    if (constraints[i].weight > EPSILON) {
      constraints[i].inserted=true;
      ins++;
    }
    i++;
  }
  writeOutput("random");
} 

void Problem::writeOutput(char* str) {
  FILE* fp = fopen(str, "w");
 
  for (int i=0; i<constraints.size(); i++) {
    Constraint & cst = constraints[i];
    if (cst.inserted) {
      assert(cst.originalWeight>=EPSILON);
      fprintf(fp, "Constraint %c%i.CB %c%i.CB %f %f %f %f\n", cst.resNameA, cst.resA, cst.resNameB, cst.resB, cst.min, cst.desired, cst.max, cst.weight);
    }
    else {
      // Use this if you want bonus constraints
      //     fprintf(fp, "Constraint %c%i.CB %c%i.CB %f %f %f %f bonus\n", cst.resNameA, cst.resA, cst.resNameB, cst.resB, cst.min, cst.desired, cst.max, cst.weight);
    }
  }
  // Use this if you want noncontacts

  /*
  for (int i=0; i<noncontacts.size(); i++) {
    Constraint & cst = noncontacts[i];
    int resA =cst.resA;
    int resB =cst.resB;
    int sep = abs(resA-resB);
    if ( sep >= 9) {
      float maxdist = (float)sep*3.8+1.6;
      float mean=8.5;
     
      // current best is 8,8.5, 10000
      fprintf(fp, "Constraint %c%i.CB %c%i.CB %f %f %f %f\n", cst.resNameA, cst.resA, cst.resNameB, cst.resB, 8.0, mean, maxdist, 0.2);
      //     fprintf(fp, "Constraint %i %i %f %f %f %f\n", cst.atomA, cst.atomB, 8.0, mean, maxdist, 0.2);
    }
  }
  */

  fclose(fp);
}

void Problem::updateScore(Constraint & cst, bool inserted) {
  assert(cst.nodeA < nodes.size());
  assert(cst.nodeB < nodes.size());
  Node & na = nodes[cst.nodeA];
  Node & nb = nodes[cst.nodeB];

  double ns = (double)nodes.size();
 
  probValue -= na.probability() + nb.probability();
 
  if (inserted) {
    na.numNeighbours++;
    nb.numNeighbours++;
  }
  else {
    na.numNeighbours--;
    nb.numNeighbours--;
    assert(na.numNeighbours >= 0);
    assert(nb.numNeighbours >= 0);
  }

  probValue += na.probability() + nb.probability();

  if (inserted) {
    weightSum += cst.weight;
    NConstraints++;
  }
  else {
    weightSum -= cst.weight;
    NConstraints--;
    assert(NConstraints >= 0);
  }
  
  avgSum = weightSum / (double) NConstraints;

  if (NConstraints == 0)
    avgSum = 0;

  score = probValue;
}

double Problem::scoreFromScratch(bool output) {
  sepCount.clear();
  sepCount.resize(1000,0);

  double sumA(0);
  for (int i=0; i<nodes.size(); i++) {
    Node & n = nodes[i];

    int N(0);
    for (int j=0; j<constraints.size(); j++) {
      Constraint & cst = constraints[j];
      if (cst.inserted && (cst.nodeA == i || cst.nodeB == i))
	N++;
    }
    n.numNeighbours = N;
    sumA += n.probability();
  }


  double sumB(0);
  int NC(0);
  for (int j=0; j<constraints.size(); j++) {
    Constraint & cst = constraints[j];
    if (cst.inserted) {
      sepCount[cst.sep]++;
      sumB += cst.weight;
      NC++;
    }
  }

  probValue = sumA;
  NConstraints = NC;
  double avg = sumB / (double)NC; 
  if (NC == 0)
    avg=0;
  
  avgSum = avg;
  weightSum = sumB;

  if (output) {
    cout << " Probability sum: " << probValue << ", CN propb: " << sumA << " weights: " << avg << " constraints: " << NC << endl; 
    
  }
  return probValue;
}
